## Langsmith agent integration

import os

__all__ = ["set_langsmith_parameters"]

def set_langsmith_parameters(langsmith_api_key: str=None,
                            langsmith_endpoint: str='https://api.smith.langchain.com',
                            langsmith_project: str='No Project',
                            langsmith_tracing: str='true') -> None:
    """
    Set up environment variables for Langsmith agent integration.

    Args:
        langsmith_api_key: API key for Langsmith
        langsmith_endpoint: Endpoint URL for Langsmith
        langsmith_project: Project name in Langsmith
        langsmith_tracing: Enable or disable tracing

    Returns:
        None
    """
    if langsmith_api_key:
        os.environ['LANGSMITH_API_KEY'] = langsmith_api_key
    if langsmith_endpoint:
        os.environ['LANGSMITH_ENDPOINT'] = langsmith_endpoint
    if langsmith_project:
        os.environ['LANGSMITH_PROJECT'] = langsmith_project
    if langsmith_tracing:
        os.environ['LANGSMITH_TRACING'] = str(langsmith_tracing)


'''
File for all tools related functions

State - Mutable data that flows through execution (e.g., messages, counters, custom fields)
Context - Immutable configuration like user IDs, session details, or application-specific configuration
Store - Persistent long-term memory across conversations
Stream Writer - Stream custom updates as tools execute
Config - RunnableConfig for the execution
Tool Call ID - ID of the current tool call
'''

from langchain.tools import tool
from typing import List, Optional, Any
from langchain_community.utilities import SQLDatabase
from mb_rag.prompts_bank import PromptManager
from langchain_core.tools import StructuredTool
from mb_rag.utils.extra import ImagePredictor
from mb_sql.sql import read_sql
from mb_sql.utils import list_schemas
from PIL import Image,ImageDraw,ImageFont
import os
import warnings
warnings.filterwarnings('ignore', message='Unable to import Axes3D')
import matplotlib.pyplot as plt
import json
from langsmith import traceable
import numpy as np

__all__ = ["list_all_tools","SQLDatabaseTools"]

def list_all_tools():
    """
    List all available tools for agents.

    Returns:
        List[str]: List of tool names.
    """
    return [
        "SQLDatabaseTools"
    ]


class SQLDatabaseTools:
    """
    Class to handle SQL Database tools.
    """
    def __init__(self, db_connection, logger=None):
        self.db_connection = db_connection
        self.logger = logger
        self.read_sql = read_sql
        self.list_schemas = list_schemas

    def _get_database_schemas(self) -> List[str]:
        """
        Get the list of schemas in the database.

        Returns:
            List[str]: List of schema names.
        """
        return self.list_schemas(self.db_connection)
    
    def to_tool_database_schemas(self):
        return StructuredTool.from_function(
            func=self._get_database_schemas,
            name="get_database_schemas",
            description="Get list of schemas in the database",
        )

    def _get_table_info(self, table_name: str, schema_name: str) -> str:
        """
        Get information about a specific table in the database.

        Args:
            table_name: Name of the table to retrieve information for.
            schema_name: Name of the schema the table belongs to.
            use_mb: Whether to use mb_sql for execution.

        Returns:
            str: Information about the table.
        """
        query = '''SELECT
                    column_name,
                    data_type,
                    character_maximum_length,
                    is_nullable,
                    column_default
                FROM information_schema.columns
                WHERE table_name = '{table_name}' AND table_schema = '{schema_name}'
                ORDER BY ordinal_position;'''.format(table_name=table_name, schema_name=schema_name)
        # if use_mb:
        return {"results": self.read_sql(query, self.db_connection)}
        # else:
        #     import pandas as pd
        #     return {"results": pd.read_csv(query, self.db_connection)}
    
    def to_tool_table_info(self):
        return StructuredTool.from_function(
            func=self._get_table_info,
            name="get_table_info",
            description="Get column info for a table",
        )

    def _get_list_tables(self, schema_name: str) -> List[str]:
        """
        Get the list of tables in a specific schema.

        Args:
            schema_name: Name of the schema to list tables from.
        Returns:
            List[str]: List of table names.
        """
        query = '''SELECT table_name
                FROM information_schema.tables
                WHERE table_schema = '{schema_name}';'''.format(schema_name=schema_name)
        results = self.read_sql(query, self.db_connection)
        return {"results": results}
    
    def to_tool_list_tables(self):
        return StructuredTool.from_function(
            func=self._get_list_tables,
            name="list_tables",
            description="List all tables in a schema",
        )

    def _base_text_to_sql(text: str = None) -> str:
        """
        Convert natural language text to a SQL query.
        
        Args:
            text: Natural language description of the desired SQL query.

        Returns:
            str: Generated SQL query.
        """
        if text:
            prompt = text
        else:
            prompt_manager = PromptManager()
            prompt = prompt_manager.get_template("SQL_AGENT_SYS_PROMPT")
        return prompt
    
    def to_tool_text_to_sql(self):
        return StructuredTool.from_function(
            func=self._base_text_to_sql,
            name="text_to_sql",
            description="Convert natural language text to SQL query",
        )

    def _execute_query_tool(self,
                            query: str, 
                            ) -> str:
        """
        Execute a SQL query on the database.
        Args:
            query: SQL query string.
            use_mb: Whether to use mb_sql for execution.

        Returns:
            str: Result of the query execution.
        """
        try:
            # if use_mb:
            results = self.read_sql(query, self.db_connection)
            results = {'results': results}
            # else:
            #     results = self.db_connection.execute_query(query)
            #     results = {'results': results}
            return results
        except Exception as e:
            return f"Error executing query: {str(e)}"
        
    def to_tool_execute_query(self):
        return StructuredTool.from_function(
            func=self._execute_query_tool,
            name="execute_query",
            description="Execute a SQL query on the database",
        )
    
class BBTools:
    """
    Class to handle Bounding Box tools.
    """
    def __init__(self, image_path: str, logger=None):
        self.image_path = image_path
        self.logger = logger

        if not os.path.exists(self.image_path):
            raise FileNotFoundError(f"Image file not found at path: {self.image_path}")

        self.image = self._load_image()

    def _load_image(self) -> Image.Image:
        """
        Load an image from the specified path.
        """
        return Image.open(self.image_path)        

    @traceable(run_type='tool',name='bounding_box_visualizer')
    def _apply_bounding_boxes(self, boxes, show: bool = False, save_location: str = './data/temp_bb_image.jpg') -> Image.Image:
        """
        Draw labeled bounding boxes on the image.
        
        Args:
            boxes: dict or JSON string:
                Expected structure: 
                {"labeled_objects": [{"label": "...", "box": [...]}, ...]}
            show: display the image using matplotlib

        Returns:
            Image.Image: image with bounding boxes and labels
        """
        self.img_bb = self.image.copy()
        draw = ImageDraw.Draw(self.img_bb)

        if isinstance(boxes, str):
            boxes_data = json.loads(boxes)
        elif isinstance(boxes, dict):
            boxes_data = boxes
        else:
            msg = "Error: Invalid boxes format received."
            if self.logger:
                self.logger.error(msg)
            else:
                print(msg)
            return self.img_bb # Return original image on error

        labeled_objects = boxes_data.get("labeled_objects", [])
        
        if not labeled_objects and isinstance(boxes_data, list):
            labeled_objects = boxes_data

        W, H = self.img_bb.size
        
        try:
            font = ImageFont.truetype("./data/arial.ttf",80)
        except IOError:
            msg = "Loading default font"
            if self.logger:
                self.logger.info(msg)
            else:
                print(msg)
            font = ImageFont.load_default()
            
        text_fill_color = "green" 
        text_color = "white"

        for obj in labeled_objects:
            if not obj.get("box") or not isinstance(obj["box"], list) or len(obj["box"]) != 4:
                continue

            label = obj["label"]
            box = obj["box"] 
            
            if 0.0 <= box[0] <= 1.0 and 0.0 <= box[2] <= 1.0:
                x0, y0, x1, y1 = (
                    round(int(box[0] * W),2),
                    round(int(box[1] * H),2),
                    round(int(box[2] * W),2),
                    round(int(box[3] * H),2)
                )
            else:
                # Assume coordinates are absolute if not normalized
                x0, y0, x1, y1 = map(int, box)
            
            x0, x1 = min(x0, x1), max(x0, x1)
            y0, y1 = min(y0, y1), max(y0, y1)

            draw.rectangle([x0, y0, x1, y1], outline="green", width=5)
            
            # try:
            bbox_text = draw.textbbox((0, 0), label, font=font)
            text_w = bbox_text[2] - bbox_text[0]
            text_h = bbox_text[3] - bbox_text[1]
            # except AttributeError:
            #     # Fallback for older PIL versions
            #     text_w, text_h = draw.textsize(label, font=font)

            text_x = x0
            text_y = max(0, y0 - text_h) 

            draw.rectangle(
                [text_x, text_y, text_x + text_w, text_y + text_h],
                fill=text_fill_color
            )
            
            draw.text((text_x, text_y), label, fill=text_color, font=font)

        if show:
            plt.imshow(self.img_bb)
            plt.axis("off")
            plt.show()

        self.img_bb.save(save_location)
        return self.img_bb
    
    def to_tool_bounding_boxes(self):
        return StructuredTool.from_function(
            func=self._apply_bounding_boxes,
            name="apply_bounding_boxes",
            description="Apply bounding boxes on image",
        )


class SEGTOOLS:
    def __init__(self,image_path : str,model_path : str, logger=None):
        
        self.image_path = image_path
        self.logger = logger

        if not os.path.exists(self.image_path):
            raise FileNotFoundError(f"Image file not found at path: {self.image_path}")

        self.image = self._load_image()

        self.predictor = ImagePredictor('./sam2_hiera_s.yaml', model_path)
        self.predictor.set_image(self.image_path)

    def _load_image(self) -> Image.Image:
        """
        Load an image from the specified path.
        """
        return Image.open(self.image_path) 
    
    @traceable(run_type='tool',name='Segmentation BB Visualizer')
    def _apply_segmentation_mask_using_bb(self,
                                          bbox_data: Optional[str], 
                                          show: bool = False, 
                                          save_location: Optional[str] = './data/temp_seg_image_bb.jpg'):
        if len(self.predictor.image.shape) == 2:
            H, W = self.predictor.image.shape
        else:
            H, W, _ = self.predictor.image.shape

        # print(f'Original bbox_data : {bbox_data}')
        bbox_data_loaded = json.loads(bbox_data) if isinstance(bbox_data, str) else bbox_data

        bbox_data = [obj["box"] for obj in bbox_data_loaded.get("labeled_objects", [])]
        bbox_labels = [obj["label"] for obj in bbox_data_loaded.get("labeled_objects", [])]
        
        # print(f'Modified bbox_data : {bbox_data}')

        bbox_data = [[bbox[1]*H, bbox[0]*W, bbox[3]*H, bbox[2]*W] for bbox in bbox_data]
        # print(f'bbox_data (pixel) : {bbox_data}')

        mask, _, _ = self.predictor.predict_item(
            bbox=bbox_data,
            labels_names=bbox_labels,
            gemini_bbox=True
        )
        # print(f'Raw mask shape: {mask.shape}')
        if len(mask.shape) == 4:
            mask = np.squeeze(mask,axis=1)
        # print(f'Squeezed mask shape: {mask.shape}')
        if mask.ndim == 4:                     # (N, C, H, W)
            mask_new = np.transpose(mask, (0, 2, 3, 1))   # (N, H, W, C)
        elif mask.ndim == 3:                   # (C, H, W)
            mask_new = np.transpose(mask, (1, 2, 0))      # (H, W, C)
        else:
            raise ValueError(f"Unexpected mask shape: {mask.shape}")

        if mask_new.shape[-1] == 1:
            mask_new = mask_new[..., 0]      

        if mask_new.ndim == 2:
            masks = [mask_new]
        elif mask_new.ndim == 3:
            if mask_new.shape[-1] > 1:
                masks = [mask_new[..., i] for i in range(mask_new.shape[-1])]
            else:
                masks = [mask_new[i] for i in range(mask_new.shape[0])]
        else:
            raise ValueError(f"Unexpected mask_new shape: {mask_new.shape}")
        
        composite = np.max(np.stack(masks, axis=0), axis=0)   # (H, W)
        if show:
            plt.imshow(composite, cmap='gray')
            plt.axis("off")
            plt.show()

        # composite = np.max(np.stack(masks, axis=0), axis=0)   # (H, W)
        img = Image.fromarray((composite * 255).astype(np.uint8))
        img.save(save_location)

    @traceable(run_type='tool',name='Segmentation Points Visualizer')
    def _apply_segmentation_mask_using_points(self,bbox_data: Optional[str],
                                              pos_points: Optional[list],
                                              neg_points: Optional[list],
                                              show: bool = False, 
                                              save_location: Optional[str] = './data/temp_seg_image_points.jpg'):
        if len(self.predictor.image.shape) == 2:
            H, W = self.predictor.image.shape
        else:
            H, W, _ = self.predictor.image.shape

        # Parse bbox_data in the same way as _apply_segmentation_mask_using_bb
        bbox_data_loaded = json.loads(bbox_data) if isinstance(bbox_data, str) else bbox_data
        bbox_list = [obj["box"] for obj in bbox_data_loaded.get("labeled_objects", [])]
        bbox_labels = [obj["label"] for obj in bbox_data_loaded.get("labeled_objects", [])]
        
        # Convert normalized coordinates to pixel coordinates
        bbox_pixel = [[bbox[1]*H, bbox[0]*W, bbox[3]*H, bbox[2]*W] for bbox in bbox_list]
        
        # Convert points to pixel coordinates (handle empty lists)
        pos_points = pos_points or []
        neg_points = neg_points or []
        pos_points_pixel = [[int(pt[0]*H), int(pt[1]*W)] for pt in pos_points] if pos_points else []
        neg_points_pixel = [[int(pt[0]*H), int(pt[1]*W)] for pt in neg_points] if neg_points else []
        
        all_points = pos_points_pixel + neg_points_pixel
        all_labels = [1]*len(pos_points_pixel) + [0]*len(neg_points_pixel)

        # Predict with bbox, points, and labels
        mask, _, _ = self.predictor.predict_item(
            bbox=bbox_pixel,
            points=all_points if all_points else None,
            point_labels=all_labels if all_labels else None,
            labels_names=bbox_labels,
            gemini_bbox=True
        )
        
        # Process mask using the same robust pattern as bb method
        if len(mask.shape) == 4:
            mask = np.squeeze(mask, axis=1)
            
        if mask.ndim == 4:                     # (N, C, H, W)
            mask_new = np.transpose(mask, (0, 2, 3, 1))   # (N, H, W, C)
        elif mask.ndim == 3:                   # (C, H, W)
            mask_new = np.transpose(mask, (1, 2, 0))      # (H, W, C)
        else:
            raise ValueError(f"Unexpected mask shape: {mask.shape}")

        if mask_new.shape[-1] == 1:
            mask_new = mask_new[..., 0]      

        if mask_new.ndim == 2:
            masks = [mask_new]
        elif mask_new.ndim == 3:
            if mask_new.shape[-1] > 1:
                masks = [mask_new[..., i] for i in range(mask_new.shape[-1])]
            else:
                masks = [mask_new[i] for i in range(mask_new.shape[0])]
        else:
            raise ValueError(f"Unexpected mask_new shape: {mask_new.shape}")
        
        composite = np.max(np.stack(masks, axis=0), axis=0)   # (H, W)
        if show:
            plt.imshow(composite, cmap='gray')
            plt.axis("off")
            plt.show()

        img = Image.fromarray((composite * 255).astype(np.uint8))
        img.save(save_location)


from sympy import re
from ..prompts_bank import PromptManager
from langchain.agents import create_agent
from .middleware import LoggingMiddleware
import base64
import os
from langgraph.graph import START, END, StateGraph,MessagesState
from .tools import SEGTOOLS,BBTools
from langsmith import traceable,uuid7
from typing import TypedDict, Optional, Dict, Any,List
import json
from langchain.agents.middleware import ModelCallLimitMiddleware,ToolCallLimitMiddleware

__all__ = ["SegmentationGraph","create_bb_agent"]

SYS_PROMPT = PromptManager().get_template("BOUNDING_BOX_LABELING_AGENT_SYS_PROMPT")

class create_bb_agent:
    """
    Create and return an AutoLabeling agent instance.
    
    Args:
        llm: The language model to use.
        langsmith_params: If True, enables LangSmith tracing.
    """
    
    def __init__(self, 
                 llm,
                langsmith_params=True,
                sys_prompt=SYS_PROMPT,
                recursion_limit: int = 3,
                user_name: str = "default_user",
                logging: bool = False,
                logger=None):

        self.llm = llm
        self.langsmith_params = langsmith_params
        self.langsmith_name = os.environ.get("LANGSMITH_PROJECT", "Seg-Labeling-Agent-Project")
        self.logging = logging
        self.logger = logger
        self.sys_prompt = sys_prompt
        self.recursion_limit = recursion_limit
        self.user_name = user_name
        self.middleware = [ModelCallLimitMiddleware(
                            run_limit=3,
                            exit_behavior="end"),
                            ToolCallLimitMiddleware(
                            tool_name="Bounding Box Visualization Tool",
                            run_limit=3,
                            exit_behavior="end")]
        if self.langsmith_params:
            self.middleware.append(LoggingMiddleware())

        self.agent = self.create_agent()

    def create_agent(self):
        """
        Create and return the AutoLabeling agent.

        Returns:
            Configured AutoLabeling agent.
        """

        @traceable(name=self.langsmith_name)
        def traced_agent():
            return create_agent(
                system_prompt=self.sys_prompt,
            tools=[],
                model=self.llm,
                middleware=self.middleware,
            ).with_config({"recursion_limit": self.recursion_limit,
                            "tags": ['bb-labeling-agent-trace'],
                            "metadata": {"user_id": self.user_name,
                                         "project": self.langsmith_name}
                        })

        return traced_agent()

    @traceable(run_type="chain", name="BB Agent Run")
    def run(self, query: str, image: str = None):
        image_base64 = self._image_to_base64(image) if image else self._image_to_base64('./temp_bb_image.jpg')

        messages = [
            {"role": "system", "content": self.sys_prompt},
            {"role": "user", "content": [
                {"type": "text", "text": query},
                {"type": "image_url", "image_url": f"data:image/jpeg;base64,{image_base64}"}
            ]}
        ]

        response = self.llm.invoke(messages,)
        # print(f'respone from LLM : {response}')
        if type(response.content)==list:
            response.content= response.content[0]['text'] ## for gemini 3 pro preview model
        raw = response.content.strip()

        if raw.startswith("```"):
            raw = raw.strip("` \n")
            if raw.startswith("json"):
                raw = raw[len("json"):].strip()
        # raw = re.sub(r"^```(?:json)?|```$", "", raw).strip()

        return raw
    
    @traceable(run_type="chain", name="Validation Segmentation Run")
    def run_seg(self, query: str, image: str = None,image_seg_bb: str = None):
        image_bb_base64 = self._image_to_base64(image) if image else self._image_to_base64('./temp_bb_image.jpg')
        image_seg_base64 = self._image_to_base64(image_seg_bb) if image_seg_bb else self._image_to_base64('./temp_seg_image_bb.jpg')

        messages = [
            {"role": "system", "content": self.sys_prompt},
            {"role": "user", "content": [
                {"type": "text", "text": query},
                {"type": "image_url", "image_url": f"data:image/jpeg;base64,{image_bb_base64}"},
                {"type": "image_url", "image_url": f"data:image/jpeg;base64,{image_seg_base64}"}
            ]}
        ]

        response = self.llm.invoke(messages,)
        # print(f'respone from LLM : {response}')
        if type(response.content)==list:
            response.content= response.content[0]['text'] ## for gemini 3 pro preview model
        raw = response.content.strip()

        if raw.startswith("```"):
            raw = raw.strip("` \n")
            if raw.startswith("json"):
                raw = raw[len("json"):].strip()
        return raw
    
    @traceable(run_type="chain", name="Validation Segmentation with Points Run")
    def run_seg_with_points(self, query: str, image: str = None, image_seg_bb: str = None, image_seg_points: str =None):
        image_bb_base64 = self._image_to_base64(image) if image else self._image_to_base64('./temp_bb_image.jpg')
        image_seg_base64 = self._image_to_base64(image_seg_bb) if image_seg_bb else self._image_to_base64('./temp_seg_image_bb.jpg')
        image_seg_points_base64 = self._image_to_base64(image_seg_points) if image_seg_points else self._image_to_base64('./temp_seg_image_points.jpg')

        messages = [
            {"role": "system", "content": self.sys_prompt},
            {"role": "user", "content": [
                {"type": "text", "text": query},
                {"type": "image_url", "image_url": f"data:image/jpeg;base64,{image_bb_base64}"},
                {"type": "image_url", "image_url": f"data:image/jpeg;base64,{image_seg_base64}"},
                {"type": "image_url", "image_url": f"data:image/jpeg;base64,{image_seg_points_base64}"}
            ]}
        ]

        response = self.llm.invoke(messages,)
        # print(f'respone from LLM : {response}')
        if type(response.content)==list:
            response.content= response.content[0]['text'] ## for gemini 3 pro preview model
        raw = response.content.strip()

        if raw.startswith("```"):
            raw = raw.strip("` \n")
            if raw.startswith("json"):
                raw = raw[len("json"):].strip()
        return raw

        
    @traceable(run_type="tool", name="Image to Base64")
    def _image_to_base64(self,image):
        """
        Convert an image file to a base64-encoded string.
        Args:
            image (str): Path to the image file.
        Returns:
            str: Base64-encoded image string.
        """
        if not os.path.exists(image):
            raise FileNotFoundError(f"Image file not found at path: {image}")

        with open(image, "rb") as f:
            return base64.b64encode(f.read()).decode('utf-8')
        

class SegmentationState(TypedDict):
    messages: List[Dict[str, Any]]
    labeled_objects: Optional[List[Dict[str, Any]]] 
    temp_bb_img_path : Optional[str]
    temp_seg_img_path : Optional[str]
    temp_segm_mask_path : Optional[str]
    temp_segm_mask_points_path : Optional[str]
    positive_points : Optional[List[int]]
    negative_points : Optional[List[int]]
    bbox_json_reason : Optional[List[str]]
    bbox_json: Optional[str]
    bb_valid: Optional[bool]
    seg_validation_reason : Optional[List[str]]
    seg_valid: Optional[bool]
    query: str
    image_path: str
    failed_labels: Optional[List[str]]
    failed_segmentation : Optional[List[str]]
    sam_model_path : str

class SegmentationGraph:
    """
    This graph uses:
    - create_segmentations_agents
    - SEGTools
    and loops until Segmentation mask is validated

    Runing :
    bb_agent = create_bb_agent(llm)
    graph = SegmentationGraph(bb_agent, "image.jpg", "Prompt")
    result = graph.run()

    print(result)
    """

    def __init__(self, agent: create_bb_agent, logger=None, show_images=False):
        self.bb_agent = agent
        self.logger = logger
        self.show_images = show_images
        self.workflow = self._build_graph()

    @traceable(run_type="chain", name="Labeler Node")
    def node_bb_labeler(self, state: SegmentationState):
            """
            Generates/Corrects the bounding box JSON based on the initial query 
            and any feedback from failed_items.
            """
            
            current_query = state["query"]
            if state.get("failed_labels"):
                failed_list = ", ".join(state["failed_labels"])
                correction_prompt = (
                    f"{current_query}\n\n"
                    f"ATTENTION: The previously generated bounding boxes for the following items were marked as incorrect or missing: **{failed_list}**. "
                    f"Please review the provided image (which shows the last attempt) and regenerate."
                )
            else:
                correction_prompt = current_query + "\n\nReturn JSON only."
                
            boxes_json = self.bb_agent.run(correction_prompt, state["image_path"])
            
            try:
                parsed_data = json.loads(boxes_json)
                labeled_objects = parsed_data.get("labeled_objects", [])
                if not isinstance(labeled_objects, list):
                    raise TypeError("Expected 'labeled_objects' to be a list.")
            except (json.JSONDecodeError, TypeError) as e:
                msg = f"Warning: LLM returned invalid JSON format: {e}. Forcing re-run."
                if self.logger:
                    self.logger.warning(msg)
                else:
                    print(msg)
                return {
                    **state, 
                    "bb_valid": False,
                    "failed_labels": ["All objects (JSON format error)"]
                }

            # msg = f"STATE in bb labeler : {state}"
            # if self.logger:
            #     self.logger.debug(msg)
            # else:
            #     print(msg)
            return {
                **state, 
                "messages": [{"role": "agent", "content": boxes_json}], 
                "bbox_json": boxes_json,
                "labeled_objects": labeled_objects, # Storing the iterable list
                "failed_labels": None
            }
    
    @traceable(run_type="tool", name="Bounding Box Visualization Tool")
    def node_bb_tool(self, state: SegmentationState):
            """Draws the bounding boxes on the image."""
            tool = BBTools(state['image_path'], logger=self.logger)
            # msg = f"STATE  in bb tool : {state}"
            # if self.logger:
            #     self.logger.debug(msg)
            # else:
            #     print(msg)
            
            # Always save, but control whether to display
            tool._apply_bounding_boxes(state["bbox_json"], show=self.show_images, save_location=state['temp_bb_img_path'])
            return state

    @traceable(run_type="llm", name="Validator Single item LLM Call")
    def _llm_validate_single_item(self, state: SegmentationState, item_data: Dict[str, Any]) -> bool:
            """
            Helper to call the LLM to validate a single item.
            The tool must draw ONLY this item's box on a temporary image.
            This is a complex step, as it requires dynamic image generation per loop iteration.
            """
            validation_prompt = f"""
            You are a Bounding Box Validator. Review the following object data and the full image with all boxes drawn on it.
            
            - **Label to Check**: {item_data['label']}
            - **Box Coordinates**: {item_data['box']}
            
            Based on the visual evidence, is the box accurate and tight?
            Your response must be a single JSON object: {{"bb_valid": True}} or {{"bb_valid": False, "reason": "..."}}.
            """
            validation_result_json = self.bb_agent.run(validation_prompt, state['temp_bb_img_path'])
            
            try:
                result = json.loads(validation_result_json)
                if result['bb_valid'] is not True:
                    state['bbox_json_reason'] = result.get("reason", "No reason provided")
                else:
                    return True
            except json.JSONDecodeError:
                msg = f"Warning: LLM returned invalid JSON format: {validation_result_json}. Forcing re-run."
                if self.logger:
                    self.logger.warning(msg)
                else:
                    print(msg)
                return False
            
    @traceable(run_type="llm", name="BB Validator LLM Call")
    def _llm_validate_full_list(self, state: SegmentationState) -> str:
        """
        Helper to call the LLM for one validation pass on the entire processed image.
        This prompt forces the LLM to return the list of items that failed.
        """
        validation_prompt = f"""
        You are a Bounding Box Validator. Review the image which contains all drawn bounding boxes.
        The objects requested were: {state['query']}
        
        You must evaluate **every single drawn bounding box** and label.
        
        Your response must be a JSON object:
        1. **"bb_valid"**: A boolean (true if ALL boxes/labels are correct, false otherwise).
        2. **"failed_labels"**: A list of strings. If "bb_valid" is true, this list is empty: []. If "bb_valid" is false, list the **names/labels** of all items that are missing, have incorrect bounding boxes, or have incorrect labels (e.g., ["blue chair", "coffee mug (missing)"]).

        Return JSON only.
        """
        return self.bb_agent.run(validation_prompt, state['temp_bb_img_path'])
    
    @traceable(run_type="chain", name="Validator Node")
    def node_bb_validator(self, state: SegmentationState):
        """
        Iterates over each item and checks its validity.
        """
        
        all_valid = True
        failed_labels = []

        for item in state.get("labeled_objects", []):
            is_valid = item.get("bb_valid", False)            
            if not is_valid:
                break 
        
        validation_result_json = self._llm_validate_full_list(state)
        
        try:
            result = json.loads(validation_result_json)
            all_valid = result.get("bb_valid", False)
            failed_labels = result.get("failed_labels", []) 
        except json.JSONDecodeError:
            all_valid = False
            failed_labels = ["All labels (Validator JSON error)"]
            
        if all_valid:
            # msg = "Validation successful."
            # if self.logger:
            #     self.logger.info(msg)
            # else:
            #     print(msg)
            return {**state, "bb_valid": True}
        else:
            msg = f"Validation failed. Items to correct: {failed_labels}"
            if self.logger:
                self.logger.warning(msg)
            else:
                print(msg)
            return {
                **state, 
                "bb_valid": False,
                "failed_labels": failed_labels 
            }

    @traceable
    def bb_route(self, state: SegmentationState):
        """
        Conditional edge: END if valid, otherwise loop back to labeler.
        """
        return "seg_tool_bb" if state["bb_valid"] else "bb_labeler"

    @traceable(run_type="tool", name="Segmentation Visualization Tool")
    def node_seg_tool_bb(self,state: SegmentationState):
        """
        Tool for get segmentation mask using SAM3 and bounding box. 
        If bounding box doesnt work it will add points to make it better
        """
        tool = SEGTOOLS(state["image_path"],state['sam_model_path'], logger=self.logger)
        # Always save, but control whether to display
        tool._apply_segmentation_mask_using_bb(state["bbox_json"], show=self.show_images, save_location=state['temp_segm_mask_path'])

    @traceable
    def node_seg_validator_bb(self,state: SegmentationState):
        """
        Check the segmentation masks
        """


        validation_prompt = f"""
            You are a Segmentation Validator. Review the segmentation mask and the original image.

            - **Label to Check**: {state['labeled_objects']}
            - The segmentation mask should not include the object's background (inside or outside) or any other objects.

            Your response must be a JSON object like:
            {{"seg_valid": True}}
            or
            {{"seg_valid": False, "reason": "...", "positive_points": [[x,y]], "negative_points": [[x,y]]}}

            Return JSON only.
            """

        validation_result_json = self.bb_agent.run_seg(
            validation_prompt,
            state["temp_bb_img_path"],
            state["temp_segm_mask_path"]
        )
        # msg = f"Validation result JSON: {validation_result_json}"
        # if self.logger:
        #     self.logger.debug(msg)
        # else:
        #     print(msg)
        try:
            result = json.loads(validation_result_json)
            # msg = f"Segmentation validation result: {result}"
            # if self.logger:
            #     self.logger.info(msg)
            # else:
            #     print(msg)
            # print(f"STATE in seg validator bb : {state}")
            return {**state, 
                    "seg_valid": result.get("seg_valid", False),
                    "seg_validation_reason": result.get("reason", ""),
                    "positive_points": result.get("positive_points", []), 
                    "negative_points": result.get("negative_points", [])}
        except json.JSONDecodeError:
            msg = f"Warning: LLM returned invalid JSON format: {validation_result_json}. Forcing re-run."
            if self.logger:
                self.logger.warning(msg)
            else:
                print(msg)
            return {**state, 
                    "seg_valid": False,
                    "seg_validation_reason": "Invalid JSON format",
                    "positive_points": [], 
                    "negative_points": []}

    @traceable(run_type="tool", name="Segmentation Visualization Tool")
    def node_seg_tool_points(self,state: SegmentationState):
        """
        Tool for get segmentation mask using SAM3 with bounding box and points.
        If bounding box doesnt work it will add points to make it better
        """
        tool = SEGTOOLS(state["image_path"],state['sam_model_path'], logger=self.logger)
        # Always save, but control whether to display
        tool._apply_segmentation_mask_using_points(
            bbox_data=state["bbox_json"],
            pos_points=state.get("positive_points", []),
            neg_points=state.get("negative_points", []),
            show=self.show_images,
            save_location=state['temp_segm_mask_points_path']
        )

    @traceable
    def seg_route1(self,state: SegmentationState):
        return END if state["seg_valid"] else "seg_tool_points"

    @traceable
    def node_seg_validation_points(self,state: SegmentationState):
        """
        Check the segmentation masks after adding points
        """
        validation_prompt = f"""
        You are a Segmentation Validator. Review the following mask data and the full image with all boxes drawn on it.
        
        - **Label to Check**: {state['labeled_objects']}

        Based on the visual evidence, is the Mask accurate and tight?
        Your response must be a single JSON object: {{"seg_valid": True}} or {{"seg_valid": False, "reason": "..."}}.
        If invalid, suggest points to add more points to improve the mask. Start by adding 1 point on either side depending upon the mask.
        {{positive_points: [[x1,y1],[x2,y2]]}} and
        {{negative_points: [[x1,y1],[x2,y2]]}}.
        Return JSON only.
        """
        validation_result_json = self.bb_agent.run_seg_with_points(validation_prompt, 
                                                              state['temp_bb_img_path'],
                                                              state['temp_segm_mask_path'],
                                                              state['temp_segm_mask_points_path'])
        
        try:
            result = json.loads(validation_result_json)
            # msg = f"Segmentation validation result with points: {result}"
            # if self.logger:
            #     self.logger.info(msg)
            # else:
            #     print(msg)
            # msg = f"STATE in seg validator points : {state}"
            # if self.logger:
            #     self.logger.debug(msg)
            # else:
            #     print(msg)
            return {**state, "seg_valid": result.get("seg_valid", False)
                    , "seg_validation_reason": result.get("reason", ""),
                    "positive_points": result.get("positive_points", []), 
                    "negative_points": result.get("negative_points", [])}
        except json.JSONDecodeError:
            msg = f"Warning: LLM returned invalid JSON format in points validation: {validation_result_json}. Forcing re-run."
            if self.logger:
                self.logger.warning(msg)
            else:
                print(msg)
            return {**state, 
                    "seg_valid": False,
                    "seg_validation_reason": "Invalid JSON format",
                    "positive_points": [], 
                    "negative_points": []}
        
    @traceable
    def seg_route2(self,state: SegmentationState):
        return END if state["seg_valid"] else "seg_tool_points"


    def _build_graph(self):
        graph = StateGraph(SegmentationState)
        graph.add_node("bb_labeler", self.node_bb_labeler)
        graph.add_node("bb_tool", self.node_bb_tool)
        graph.add_node("bb_validator", self.node_bb_validator)

        graph.add_node("seg_tool_bb", self.node_seg_tool_bb)
        graph.add_node("seg_tool_bb_validation",self.node_seg_validator_bb)

        graph.add_node("seg_tool_points",self.node_seg_tool_points)
        graph.add_node("seg_tool_points_validation",self.node_seg_validation_points)


        graph.add_edge(START, "bb_labeler")
        graph.add_edge("bb_labeler", "bb_tool")
        graph.add_edge("bb_tool", "bb_validator")

        graph.add_conditional_edges(
            "bb_validator", 
            self.bb_route, 
            {
                "seg_tool_bb": "seg_tool_bb",
                "bb_labeler": "bb_labeler"
            }
        )

        graph.add_edge("seg_tool_bb", "seg_tool_bb_validation")

        graph.add_conditional_edges(
            "seg_tool_bb_validation",
            self.seg_route1,
            {
                END: END,
                "seg_tool_points": "seg_tool_points"
            }
        )

        graph.add_edge("seg_tool_points", "seg_tool_points_validation")

        graph.add_conditional_edges(
            "seg_tool_points_validation",
            self.seg_route2,
            {
                END: END,
                "seg_tool_points": "seg_tool_points"
            }
        )
        return graph.compile()

    @traceable
    def run(self, image_path: str, 
            query: str, 
            temp_image :str = './data/temp_bb_image.jpg', 
            temp_segm_mask_path: str = './data/temp_seg_image_bb.jpg',
            temp_segm_mask_points_path: str = './data/temp_seg_image_points.jpg',
            sam_model_path: str = './models/sam2_hiera_small.pt',
            recursion_limit: int = 3):
        self.image_path = image_path
        self.query = query
        self.temp_image = temp_image
        self.temp_segm_mask_path = temp_segm_mask_path
        self.temp_segm_mask_points_path = temp_segm_mask_points_path
        self.sam_model_path = sam_model_path
        return self.workflow.invoke(
            {
                "agent": self.bb_agent,
                "image_path": self.image_path,
                "query": self.query,
                "temp_bb_img_path": self.temp_image,
                "temp_segm_mask_path": self.temp_segm_mask_path,
                "temp_segm_mask_points_path": self.temp_segm_mask_points_path,
                "bb_valid": False,
                "bb_validation_reason": "",
                "seg_valid": False,
                "sam_model_path": self.sam_model_path,
                "seg_validation_reason": "",
                "failed_labels": [],
                "failed_segmentation": [],
                "positive_points": [],
                "negative_points": []
            },
            config={"recursion_limit": recursion_limit}
        )
    


## middleware / hooks for agents
'''
Middleware Classes:

ContextEditingMiddleware:	Edit context in agent workflows
HumanInTheLoopMiddleware:	Enable human intervention in agent processes
LLMToolSelectorMiddleware:	Select tools using LLM-based logic
LLMToolEmulator:	        Emulate tool execution using LLM
ModelCallLimitMiddleware:	Limit the number of model calls
ModelFallbackMiddleware:	Provide fallback model options
PIIMiddleware:	            Handle personally identifiable information
SummarizationMiddleware: 	Summarize content in agent workflows
TodoListMiddleware:      	Manage to-do lists in agent processes
ToolCallLimitMiddleware:    Limit the number of tool calls
AgentMiddleware:	        Base middleware class for agent operations


Components for middleware functionalities in AI agents.

AgentState:             Function to clear tool usage edits
InterruptOnConfig:      Configuration class for interruption handling
ModelRequest:           Represent a request to a model
ModelResponse:          Represent a response from a model
before_model:           Function executed before model calls
after_model:            Function executed after model calls
wrap_model_call:        Function wrapper for model calls
wrap_tool_call:         Function wrapper for tool calls


'''
from typing import List,Optional,Dict,Any
import time
from langchain.agents.middleware import AgentMiddleware,AgentState,ModelRequest,ModelResponse
from langchain.messages import AIMessage
from langgraph.runtime import Runtime
import json

__all__ = [
    "SQLGuardRailsMiddleware",
    "timing_middleware",
    "LoggingMiddleware",
]
        
def timing_middleware(timer: int, state: AgentState, logger=None) -> str:
    """
    Middleware to log the time taken for agent execution.
    Args:
        timer (int): The maximum time to wait for the agent to respond.
        state (AgentState): The state of the agent.
        logger: Optional logger instance.
    Returns:
        str: The original input.
    """
    start_time = time.time()
    result = state.get("input", "")
    end_time = time.time()
    elapsed_time = end_time - start_time
    msg = f"[TimingMiddleware] Agent execution time: {elapsed_time:.4f} seconds"
    if logger:
        logger.info(msg)
    else:
        print(msg)
    return result

class SQLGuardRailsMiddleware(AgentMiddleware):
    """
    Middleware to prevent SQL Modifications.
    """
    def __init__(self, logger=None):
        super().__init__()
        self.logger = logger
        
    def _extract_sql_from_state(self, state: AgentState) -> str:
        messages = state.get("messages", [])
        for msg in reversed(messages):
            if isinstance(msg, AIMessage):
                fn_call = msg.additional_kwargs.get("function_call", {})
                if fn_call and fn_call.get("name") == "sql_db_query":
                    try:
                        args = json.loads(fn_call.get("arguments", "{}"))
                        return args.get("query", "")
                    except Exception as e:
                        msg = f"[SQLGuardRailsMiddleware] Error parsing SQL args: {e}"
                        if self.logger:
                            self.logger.error(msg)
                        else:
                            print(msg)
        return ""
    
    def after_model(self, state: AgentState, runtime: Runtime) -> Dict[str, Any]:
        input_query = self._extract_sql_from_state(state)
        query_upper = input_query.upper()
        if any(op in query_upper for op in ["UPDATE", "DELETE", "INSERT", "DROP", "ALTER", "CREATE"]):
            msg = f"[SQLGuardRailsMiddleware] SQL Table modification access detected: {input_query}"
            if self.logger:
                self.logger.warning(msg)
            else:
                print(msg)
            return {
                    "messages": [AIMessage("I cannot respond to that request as it involves modifying SQL tables.")],
                    "jump_to": "end"
            }
        return None
        

class LoggingMiddleware(AgentMiddleware):
    """
    Middleware to log agent interactions.
    """
    def __init__(self, logger=None):
        super().__init__()
        self.logger = logger
        
    def before_model(self,state: AgentState) -> None:
        msg = f"[LoggingMiddleware] Before model call with input: {state}"
        if self.logger:
            self.logger.debug(msg)
        else:
            print(msg)
        return None

    def after_model(self, state: AgentState) -> None:
        msg = f"[LoggingMiddleware] After model call with output: {state}"
        if self.logger:
            self.logger.debug(msg)
        else:
            print(msg)
        return None
    
    def after_agent(self, state: AgentState) -> None:
        msg = f"[LoggingMiddleware] After agent execution with final state: {state}"
        if self.logger:
            self.logger.debug(msg)
        else:
            print(msg)
        return None
    










# class SQLGuardRailsMiddleware(AgentMiddleware):
#     """
#     Middleware to prevent SQL Modifications.
#     """
#     def _extract_sql_from_state(self, state: AgentState) -> str:
#         messages = state.get("messages", [])
#         for msg in reversed(messages):
#             if isinstance(msg, AIMessage):
#                 fn_call = msg.additional_kwargs.get("function_call", {})
#                 if fn_call and fn_call.get("name") == "sql_db_query":
#                     try:
#                         args = json.loads(fn_call.get("arguments", "{}"))
#                         return args.get("query", "")
#                     except Exception as e:
#                         print(f"[SQLGuardRailsMiddleware] Error parsing SQL args: {e}")
#         return ""
    
#     def after_model(self, state: AgentState, runtime: Runtime) -> Dict[str, Any]:
#         input_query = self._extract_sql_from_state(state)
#         query_upper = input_query.upper()
#         if any(op in query_upper for op in ["UPDATE", "DELETE", "INSERT", "DROP", "ALTER", "CREATE"]):
#             print(f"[SQLGuardRailsMiddleware] SQL Table modification access detected: {input_query}")
#             return {
#                     "messages": [AIMessage("I cannot respond to that request as it involves modifying SQL tables.")],
#                     "jump_to": "end"
#             }
#         return None


"""
Base Agent runner implementation
"""

from typing import List,Optional
from mb_rag.utils import viewer
from mb_rag.agents import get_langsmith
from mb_rag.basic import ModelFactory
import os 
from mb_rag.agents.tools import list_all_tools

__all__ = ["AgentRunner"]


class AgentDict:
    """
    Base class to run AI agents 
    
    Attributes:
        llm : Language model instance. (ModelFactory : mb_rag.basic.ModelFactory)
        tools (List[str]) : List of available tools (Check mb_rag.agents.tools for available tools)
        agent_instance (str): The agent instance
        langsmith_params (dict): Langsmith parameters if any. {langsmith_api_key, langsmith_endpoint, langsmith_project, langsmith_tracing}
        middleware (List[str]): Middleware for the agent
        extra_params (dict): Any extra parameters
    """
    def __init__(self,
                llm : ModelFactory,
                agent_instance : str,
                tools: Optional[List[str]] = None,
                langsmith_params: Optional[dict] = None,
                middleware: Optional[List[str]] = None,
                extra_params: Optional[dict] = None,
                logger=None):

        self.llm = llm
        self.tools = tools if tools else []
        self.agent_instance = agent_instance if agent_instance else None
        self.langsmith_params = langsmith_params if langsmith_params else {}
        self.middleware = middleware if middleware else []
        self.extra_params = extra_params if extra_params else {}
        self.logger = logger

    def _supervise_agent(self, input_query: str) -> str:
        """
        Supervise the agent's execution with the given input query.

        Args:
            input_query (str): The input query for the agent.

        Returns:
            str: The agent's response.
        """
        try:
            if self.langsmith_params:
                self._load_langsmith()
            agent = self.create_main_agent()
            response = agent.run(input_query)
            viewer.display_response(response)
            return response
        except Exception as e:
            viewer.display_error(e)
            msg = f"[Agent Error] {e}"
            if self.logger:
                self.logger.error(msg)
            else:
                print(msg)
            return str(e)

    def _support_agent(self, input_query: str) -> str:
        """
        Support the agent's execution with the given input query.

        Args:
            input_query (str): The input query for the agent.

        Returns:
            str: The agent's response.
        """
        try:
            if self.langsmith_params:
                self._load_langsmith()
            agent = self.create_main_agent()
            response = agent.run(input_query)
            viewer.display_response(response)
            return response
        except Exception as e:
            viewer.display_error(e)
            msg = f"[Agent Error] {e}"
            if self.logger:
                self.logger.error(msg)
            else:
                print(msg)
            return str(e)

    def create_main_agent(self):
        """
        Create and configure the main agent.
        
        Returns:
            Configured main agent.
        """
        pass  # To be implemented in subclasses

    def _get_all_tools(self):
        """
        Retrieve all available tools for the agent.
        
        Returns:
            List of tools.
        """
        return self.tools
    
    def _list_all_tools(self):
        """
        List all available tools for the agent.
        
        Returns:
            List of tool names.
        """
        return list_all_tools()

    def _get_all_middlewares(self):
        """
        Retrieve all middlewares for the agent.
        
        Returns:
            List of middlewares.
        """
        return self.middleware
    
    def _get_all_extra_params(self):
        """
        Retrieve all extra parameters for the agent.
        
        Returns:
            Dictionary of extra parameters.
        """
        return self.extra_params
    
    def _get_langsmith_params(self):
        """
        Retrieve Langsmith parameters for the agent.
        
        Returns:
            Dictionary of Langsmith parameters.
        """
        return self.langsmith_params
    
    def _load_langsmith(self):
        """
        Load Langsmith parameters into environment variables.
        
        Returns:
            None
        """
        if os.getenv('LANGSMITH_API_KEY') is not None:
                langsmith_api_key=self.langsmith_params.get('langsmith_api_key', None)
        get_langsmith.set_langsmith_parameters(
            langsmith_api_key=langsmith_api_key,
            langsmith_endpoint=self.langsmith_params.get('langsmith_endpoint', 'https://api.smith.langchain.com'),
            langsmith_project=self.langsmith_params.get('langsmith_project', 'No Project'),
            langsmith_tracing=self.langsmith_params.get('langsmith_tracing', 'false')
        )
        msg = "Langsmith parameters: {} loaded into environment variables.".format(self.langsmith_params)
        if self.logger:
            self.logger.info(msg)
        else:
            print(msg)
        return None
        if self.logger:
            self.logger.info(msg)
        else:
            print(msg)
        return None
    



## extra functions for RAG's

import os
from dotenv import load_dotenv
import importlib.util
from typing import Optional,List,Dict,Union,Tuple
import cv2
import numpy as np
import matplotlib.pyplot as plt


__all__ = ["load_env_file", "check_package", "pdf_to_text", "convert_pdfs_in_folder","SAM2Processor","ImagePredictor"]

def load_env_file(file_path='.env'):
    """
    Load environment variables from a .env file.

    Args:
        file_path (str): Path to the .env file. Defaults to '.env'.

    Returns:
        None
    """
    load_dotenv(file_path)

    # Get the loaded environment variables
    env_vars = os.environ

    return env_vars

def check_package(package_name, error_message=None):
    """
    Check if a package is installed
    Args:
        package_name (str): Name of the package
        error_message (str, optional): Custom error message to display if the package is not found
    Returns:
        bool: True if package is installed, False otherwise
    """

    if importlib.util.find_spec(package_name) is not None:
        return True
    else:
        if error_message:
            print(error_message)
        else:
            print(f"Package '{package_name}' not found.")
    return False

def pdf_to_text(pdf_path):
    """Extract text from a PDF file."""
    text = ""
    try:
        if not check_package("PyPDF2"):
            raise ImportError("PyPDF2 package not found. Please install it using: pip install pypdf2")
        import PyPDF2
        with open(pdf_path, "rb") as file:
            reader = PyPDF2.PdfReader(file)
            for page in reader.pages:
                text += page.extract_text() + "\n"
    except PyPDF2.errors.PdfReadError as e:
        print(f"Error reading {pdf_path}: {e}")
    except Exception as e:
        print(f"An unexpected error occurred with {pdf_path}: {e}")
    return text

def convert_pdfs_in_folder(folder_path):
    """
    Convert all PDF files in the given folder to text files.
    Args:
        folder_path (str): Path to the folder containing the PDF files.
    Returns:
        None
    Example : convert_pdfs_in_folder('/folder_path') # folder_path is the path to the folder containing the PDF files.
    The converted PDF files and text files will be created in the same folder
    """
    for filename in os.listdir(folder_path):
        if filename.endswith('.pdf'):
            pdf_path = os.path.join(folder_path, filename)
            text = pdf_to_text(pdf_path)
            if text:  # Only write to file if text is not empty
                text_filename = os.path.splitext(filename)[0] + '.txt'
                text_path = os.path.join(folder_path, text_filename)
                with open(text_path, 'w', encoding='utf-8') as text_file:
                    text_file.write(text)
                print(f"Converted: {filename} to {text_filename}")
            else:
                print(f"No text extracted from {filename}")


class SAM2Processor:
    """
    Main class for SAM2 model operations including mask generation and visualization.
    """
    
    def __init__(self, sam2_checkpoint: str = '../checkpoints/sam2_hiera_large.pt',
                 model_cfg: str = 'sam2_hiera_l.yaml', device: str = 'cpu'):
        """
        Initialize SAM2Processor.

        Args:
            sam2_checkpoint (str): Path to model checkpoint
            model_cfg (str): Path to model configuration
            device (str): Device to run on
        """
        self.device = device
        self.sam2_checkpoint = sam2_checkpoint
        self.model_cfg = model_cfg
        self.mask_generator = self._initialize_mask_generator()


    def _initialize_mask_generator(self):
        """Initialize the mask generator."""
        from sam2.build_sam import build_sam2
        from sam2.automatic_mask_generator import SAM2AutomaticMaskGenerator
        sam2 = build_sam2(self.model_cfg, self.sam2_checkpoint, 
                         device=self.device, apply_postprocessing=False)
        return SAM2AutomaticMaskGenerator(sam2)

    def show_anns(self, anns: List[Dict], borders: bool = True, show: bool = True) -> Optional[np.ndarray]:
        """Display annotations on an image."""
        if len(anns) == 0:
            return
        sorted_anns = sorted(anns, key=(lambda x: x['area']), reverse=True)
        ax = plt.gca()
        ax.set_autoscale_on(False)
     
        img = np.ones((sorted_anns[0]['segmentation'].shape[0], 
                      sorted_anns[0]['segmentation'].shape[1], 4))
        img[:,:,3] = 0
        for ann in sorted_anns:
            m = ann['segmentation']
            color_mask = np.concatenate([np.random.random(3), [0.5]])
            img[m] = color_mask 
            if borders:
                contours, _ = cv2.findContours(m.astype(np.uint8),
                                             cv2.RETR_EXTERNAL, 
                                             cv2.CHAIN_APPROX_NONE)
                contours = [cv2.approxPolyDP(contour, epsilon=0.01, closed=True) 
                          for contour in contours]
                cv2.drawContours(img, contours, -1, (0,0,1,0.4), thickness=1)
        if show:    
            ax.imshow(img)
        return img

    def get_similarity_value(self, box1: List[float], box2: List[float]) -> float:
        """Calculate similarity between two bounding boxes."""
        val1 = abs(box1[0]-box2[0])
        val2 = abs(box1[1]-box2[1])
        val3 = abs(box1[2]-box2[2])
        val4 = abs(box1[3]-box2[3])
        return val1 + val2 + val3 + val4

    def get_final_similar_box(self, box1: List[float], box2: List[List[float]]) -> Tuple[List[float], int]:
        """Find the most similar box from a list of boxes."""
        best_box = None
        best_val = None
        index = None
        for i in box2:
            val = self.get_similarity_value(box1, i)
            if best_box is None or val < best_val:
                best_box = i
                best_val = val
                index = box2.index(i)
        return best_box, index

    def get_mask_for_bbox(self, image_path: str, bbox_value: List[float],
                         show_full: bool = False, show_final: bool = False) -> Tuple[np.ndarray, List[float], List[List[float]]]:
        """Get mask for a specific bounding box."""
        print('Getting mask')
        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        mask_full = self.mask_generator.generate(image)
        if show_full:
            self.show_anns(mask_full)
        print('Getting final mask')

        main_bbox = []
        for i in mask_full:
            mask_val = [i['bbox'][1], i['bbox'][0],
                       (i['bbox'][3]+i['bbox'][1]), (i['bbox'][2]+i['bbox'][0])]
            main_bbox.append(mask_val)

        value_list, index = self.get_final_similar_box(bbox_value, main_bbox)
        final_mask = mask_full[index]
        final_bbox = [final_mask['bbox'][1], final_mask['bbox'][0],
                     (final_mask['bbox'][3]+final_mask['bbox'][1]),
                     (final_mask['bbox'][2]+final_mask['bbox'][0])]
        if show_final:
            self.show_anns([final_mask])
        return final_mask['segmentation'], final_bbox, main_bbox

    def get_all_masks(self, image_path: str) -> List[Dict]:
        """Get all masks for an image."""
        print('Getting all masks')
        image = cv2.imread(image_path)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        mask_full = self.mask_generator.generate(image)
        print('Getting final mask')
        return mask_full

    @staticmethod
    def show_mask(mask: np.ndarray, ax: plt.Axes, obj_id: Optional[int] = None,
                 random_color: bool = False) -> None:
        """Display a mask on a given axis."""
        if random_color:
            color = np.concatenate([np.random.random(3), np.array([0.6])], axis=0)
        else:
            cmap = plt.get_cmap("tab10")
            cmap_idx = 0 if obj_id is None else obj_id
            color = np.array([*cmap(cmap_idx)[:3], 0.6])
        h, w = mask.shape[-2:]
        mask_image = mask.reshape(h, w, 1) * color.reshape(1, 1, -1)
        ax.imshow(mask_image)

    @staticmethod
    def show_masks_image(
        image: np.ndarray,
        masks: List[np.ndarray],
        scores: List[float],
        point_coords: Optional[np.ndarray] = None,
        box_coords: Optional[Union[List, np.ndarray]] = None,
        point_labels: Optional[np.ndarray] = None,
        labels_names: Optional[List[str]] = None,
        borders: bool = True
    ) -> None:
        """Display multiple masks on an image."""

        if box_coords is not None:
            box_coords = np.array(box_coords, dtype=np.float32)

            if box_coords.ndim == 1:
                box_coords = box_coords.reshape(1, 4)

            if box_coords.shape[0] == 1 and len(masks) > 1:
                box_coords = np.repeat(box_coords, len(masks), axis=0)

            per_mask_boxes = (box_coords.shape[0] == len(masks))
        else:
            per_mask_boxes = False

        plt.figure(figsize=(10, 10))
        ax = plt.gca()
        ax.imshow(image)

        masks = np.stack(masks)
        for i, mask in enumerate(masks):
            SAM2Processor.show_mask(mask, ax, obj_id=i)

        if point_coords is not None:
            assert point_labels is not None
            SAM2Processor.show_points(point_coords, point_labels, ax)

        if box_coords is not None:
            if per_mask_boxes:
                SAM2Processor.show_box(box_coords, ax, labels=labels_names)
            else:
                SAM2Processor.show_box(box_coords, ax, labels=labels_names)

        # if len(scores) > 1:
            # ax.set_title("Masks: " + ", ".join(str(i+1) for i in range(len(masks))), fontsize=18)

        ax.axis("off")
        plt.show()

    @staticmethod
    def show_points(coords: np.ndarray,
                    labels: np.ndarray,
                    ax: plt.Axes,
                    marker_size: int = 200) -> None:
        """
        Draw 2D points on matplotlib axis.
        
        Supports:
            coords: (P, 2) or (B, P, 2)
            labels: (P,) or (B, P)
        
        Positive label (1) = green star
        Negative label (0) = red star
        """

        coords = np.asarray(coords)
        labels = np.asarray(labels)

        if coords.ndim == 2:
            pos = coords[labels == 1]
            neg = coords[labels == 0]

            if len(pos) > 0:
                ax.scatter(pos[:, 0], pos[:, 1],
                        color='green', marker='*',
                        s=marker_size, edgecolor='white', linewidth=1.2)

            if len(neg) > 0:
                ax.scatter(neg[:, 0], neg[:, 1],
                        color='red', marker='*',
                        s=marker_size, edgecolor='white', linewidth=1.2)
            return

        if coords.ndim == 3:
            B = coords.shape[0]

            for i in range(B):
                pts = coords[i]
                lbl = labels[i]

                pos = pts[lbl == 1]
                neg = pts[lbl == 0]

                if len(pos) > 0:
                    ax.scatter(pos[:, 0], pos[:, 1],
                            color='green', marker='*',
                            s=marker_size, edgecolor='white', linewidth=1.2)

                if len(neg) > 0:
                    ax.scatter(neg[:, 0], neg[:, 1],
                            color='red', marker='*',
                            s=marker_size, edgecolor='white', linewidth=1.2)
            return

        raise ValueError(f"coords must be shape (P,2) or (B,P,2). Got {coords.shape}")

        
    @staticmethod
    def show_box(box: Union[List, np.ndarray],
                  ax: plt.Axes,
                  labels: Optional[List[str]] = None) -> None:
        """Display one or multiple bounding boxes on an axis."""
        
        box = np.array(box, dtype=np.float32)

        if box.ndim == 1:
            box = box.reshape(1, 4)

        if labels is not None:
            if len(labels) == 1 and len(box) > 1:
                labels = labels * len(box)
            elif len(labels) != len(box):
                labels = labels + [""] * (len(box) - len(labels))
        else:
            labels = [""] * len(box)

        for (x1, y1, x2, y2), label in zip(box, labels):
            w, h = x2 - x1, y2 - y1

            ax.add_patch(
                plt.Rectangle(
                    (x1, y1),
                    w,
                    h,
                    edgecolor='green',
                    facecolor=(0, 0, 0, 0),
                    lw=2
                )
            )

            if label:
                ax.text(
                    x1,
                    y1 - 5,
                    label,
                    fontsize=12,
                    color='green',
                    bbox=dict(
                        facecolor='black',
                        alpha=0.5,
                        edgecolor='none',
                        pad=2
                    )
                )

class ImagePredictor:
    """Class for image prediction using SAM2."""

    def __init__(self, model_cfg: str, sam2_checkpoint: str, device: str = 'cpu'):
        """Initialize ImagePredictor."""
        from sam2.sam2_image_predictor import SAM2ImagePredictor
        from sam2.build_sam import build_sam2

        self.predictor = SAM2ImagePredictor(build_sam2(model_cfg, sam2_checkpoint, device=device))
        self.image = None

    def set_image(self, image: Union[str, np.ndarray]) -> None:
        """Set the image for prediction."""
        if isinstance(image, str):
            image = cv2.imread(image)
            self.image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        else:
            self.image = image
        self.predictor.set_image(self.image)

    def predict_item(self, 
                    bbox: Optional[List[List[float]]] = None,
                    points: Optional[List[List[float]]] = None,
                    point_labels: Optional[List[int]] = None,
                    labels_names: Optional[List[str]] = None,
                    show: bool = True, 
                    gemini_bbox: bool = True,
                    **kwargs):

        all_masks = []
        all_scores = []
        all_logits = []
        all_boxes = []
        all_points = []
        all_labels = []

        if bbox is not None:
            bbox = np.array(bbox, dtype=np.float32)
            if gemini_bbox:
                bbox = bbox[:, [1, 0, 3, 2]]   # reorder
        else:
            bbox = []

        for i, single_box in enumerate(bbox if len(bbox) > 0 else [None]):

            predict_args = {}

            if points is not None and point_labels is not None:
                predict_args["point_coords"] = np.array(points, dtype=np.float32)
                predict_args["point_labels"] = np.array(point_labels, np.int32)
                all_points.append(np.array(points, dtype=np.float32))
                all_labels.append(np.array(point_labels, np.int32))

            if single_box is not None:
                predict_args["box"] = single_box.astype(np.float32)
                all_boxes.append(single_box.astype(np.float32))

            masks, scores, logits = self.predictor.predict(
                **predict_args, multimask_output=False, **kwargs
            )

            sorted_ind = np.argsort(scores)[::-1]
            masks = masks[sorted_ind]
            scores = scores[sorted_ind]
            logits = logits[sorted_ind]

            all_masks.append(masks[0])
            all_scores.append(scores[0])
            all_logits.append(logits[0])

        all_masks = np.array(all_masks)
        all_scores = np.array(all_scores)
        all_logits = np.array(all_logits)
        all_boxes = np.array(all_boxes)
        if len(all_points)==0:
            all_points = None
            all_labels = None
        else:
            all_points = np.array(all_points)
            all_labels = np.array(all_labels)

        if show:
            self._visualize_prediction(all_masks, all_scores, all_points, all_boxes, all_labels, labels_names)

        return all_masks, all_scores, all_logits
    
    def _visualize_prediction(self, masks: np.ndarray, scores: np.ndarray,points: Optional[np.ndarray] = None,
                            bbox: Optional[np.ndarray] = None, point_labels: Optional[np.ndarray] = None,
                            labels_names: Optional[List[str]] = None) -> None:
        """Visualize the prediction."""
        # plt.figure(figsize=(10, 10))
        # plt.imshow(self.image)
        SAM2Processor.show_masks_image(self.image, masks, scores, point_coords=points,
                                     box_coords=bbox, point_labels=point_labels, labels_names=labels_names)


from langchain_core.messages import AIMessage, HumanMessage, SystemMessage
from typing import Optional, List, Any, Union

__all__ = [
    'ConversationModel'
]

class ConversationModel:
    """
    A class to handle conversation with AI models
    
    Attributes:
        chatbot: The AI model for conversation
        message_list (List): List of conversation messages
        file_path (str): Path to save/load conversations. Can be local or S3
    """
    
    def __init__(self, 
                llm: Any,
                message_list: Optional[List[Any]] = None,
                file_path: Optional[str] = None,
                **kwargs) -> None:
        """Initialize conversation model"""
        self.chatbot = llm
        if message_list:
            self.message_list = message_list
        else:
            self.message_list = []
        if file_path:
            self.file_path = file_path
        else:
            self.file_path = None

    def initialize_conversation(self,context_message: str = "") -> None:
        """Initialize conversation state.
            Getting the content from file_path if provided"""
        if self.file_path:
            self.load_conversation()

        if context_message:
            self.message_list.append(SystemMessage(content=context_message))
        else:
            self.message_list.append(SystemMessage(content="""This is conversation model. 
                            Look into the conversation history and answer the question if provided.
                            Give a brief introduction of the conversation history."""))
        message_list_content = "".join(self.all_messages_content)
        return self.add_message(message_list_content,get_content_only=True)    

    def _ask_question(self,query: str,images: list = None, 
                     get_content_only: bool = True) -> str:
        """
        Ask a question and get response
        Args:
            query: Question to ask
            get_content_only: Whether to return only content
        Returns:
            str: Response from the model
        """
        if images:
            res = self.chatbot.invoke_query(query,images=images,get_content_only=get_content_only)
        else:
            res = self.chatbot.invoke_query(query,get_content_only=get_content_only)
        return res

    def add_message(self, query: str,images: list = None,get_content_only: bool = True) -> str:
        """
        Add a message to the conversation
        Args:
            query (str): Question to ask
            images (list): List of images to send to the model
            get_content_only (bool): Whether to return only content
        Returns:
            str: Response from the chatbot
        """
        self.message_list.append(HumanMessage(content=query))
        res = self._ask_question(query,images=images,get_content_only=get_content_only)
        self.message_list.append(AIMessage(content=res))
        return res

    @property
    def all_messages(self) -> List[Union[SystemMessage, HumanMessage, AIMessage]]:
        """Get all messages"""
        return self.message_list

    @property
    def last_message(self) -> str:
        """Get the last message"""
        return self.message_list[-1].content

    @property
    def all_messages_content(self) -> List[str]:
        """Get content of all messages"""
        return [message.content for message in self.message_list]

    def _is_s3_path(self, path: str) -> bool:
        """
        Check if path is an S3 path
        Args:
            path (str): Path to check
        Returns:
            bool: True if S3 path
        """
        return path.startswith("s3://")

    def save_conversation(self, file_path: Optional[str] = None, **kwargs) -> bool:
        """
        Save the conversation
        Args:
            file_path: Path to save the conversation
            **kwargs: Additional arguments for S3
        Returns:
            bool: Success status
        """
        if self._is_s3_path(file_path or self.file_path):
            print("Saving conversation to S3.")
            self.save_file_path = file_path
            return self._save_to_s3(self.file_path,**kwargs)
        return self._save_to_file(file_path or self.file_path)

    def _save_to_s3(self,**kwargs) -> bool:
        """Save conversation to S3"""
        try:
            client = kwargs.get('client', self.client)
            bucket = kwargs.get('bucket', self.bucket)
            client.put_object(
                Body=str(self.message_list),
                Bucket=bucket,
                Key=self.save_file_path
            )
            print(f"Conversation saved to s3_path: {self.s3_path}")
            return True
        except Exception as e:
            raise ValueError(f"Error saving conversation to s3: {e}")

    def _save_to_file(self, file_path: str) -> bool:
        """Save conversation to file"""
        try:
            with open(file_path, 'w') as f:
                for message in self.message_list:
                    f.write(f"{message.content}\n")
            print(f"Conversation saved to file: {file_path}")
            return True
        except Exception as e:
            raise ValueError(f"Error saving conversation to file: {e}")

    def load_conversation(self, file_path: Optional[str] = None, **kwargs) -> List[Any]:
        """
        Load a conversation
        Args:
            file_path: Path to load from
            **kwargs: Additional arguments for S3
        Returns:
            List: Loaded messages
        """
        self.message_list = []
        if self._is_s3_path(file_path or self.file_path):
            print("Loading conversation from S3.")
            self.file_path = file_path
            return self._load_from_s3(**kwargs)
        return self._load_from_file(file_path or self.file_path)

    def _load_from_s3(self, **kwargs) -> List[Any]:
        """Load conversation from S3"""
        try:
            client = kwargs.get('client', self.client)
            bucket = kwargs.get('bucket', self.bucket)
            res = client.get_response(client, bucket, self.s3_path)
            res_str = eval(res['Body'].read().decode('utf-8'))
            self.message_list = [SystemMessage(content=res_str)]
            print(f"Conversation loaded from s3_path: {self.file_path}")
            return self.message_list
        except Exception as e:
            raise ValueError(f"Error loading conversation from s3: {e}")

    def _load_from_file(self, file_path: str) -> List[Any]:
        """Load conversation from file"""            
        try:
            with open(file_path, 'r') as f:
                lines = f.readlines()
            for line in lines:
                self.message_list.append(SystemMessage(content=line))
            print(f"Conversation loaded from file: {file_path}")
            return self.message_list
        except Exception as e:
            raise ValueError(f"Error loading conversation from file: {e}")



from ..prompts_bank import PromptManager
from langchain.agents import create_agent
from .middleware import LoggingMiddleware
import base64
import os
from langgraph.graph import START, END, StateGraph,MessagesState
from .tools import BBTools
from langsmith import traceable
from typing import TypedDict, Optional, Dict, Any,List
import json
from langchain.agents.middleware import ModelCallLimitMiddleware,ToolCallLimitMiddleware


__all__ = ["create_labeling_agent","LabelingGraph"]

SYS_PROMPT = PromptManager().get_template("BOUNDING_BOX_LABELING_AGENT_SYS_PROMPT")

class create_labeling_agent:
    """
    Create and return an AutoLabeling agent instance.
    
    Args:
        llm: The language model to use.
        langsmith_params: If True, enables LangSmith tracing.
    """
    
    def __init__(self, 
                 llm,
                langsmith_params=True,
                sys_prompt=SYS_PROMPT,
                recursion_limit: int = 50,
                user_name: str = "default_user",
                logging: bool = False,
                logger=None):

        self.llm = llm
        self.langsmith_params = langsmith_params
        self.langsmith_name = os.environ.get("LANGSMITH_PROJECT", "BB-Labeling-Agent-Project")
        self.logging = logging
        self.logger = logger
        self.sys_prompt = sys_prompt
        self.recursion_limit = recursion_limit
        self.user_name = user_name
        self.middleware = [ModelCallLimitMiddleware(
                            run_limit=3,
                            exit_behavior="end"),
                            ToolCallLimitMiddleware(
                            tool_name="Bounding Box Visualization Tool",
                            run_limit=3,
                            exit_behavior="end")]
        if self.langsmith_params:
            self.middleware.append(LoggingMiddleware())

        self.agent = self.create_agent()

    def create_agent(self):
        """
        Create and return the AutoLabeling agent.

        Returns:
            Configured AutoLabeling agent.
        """

        @traceable(name=self.langsmith_name)
        def traced_agent():
            return create_agent(
                system_prompt=self.sys_prompt,
            tools=[],
                model=self.llm,
                middleware=self.middleware,
            ).with_config({"recursion_limit": self.recursion_limit,
                            "tags": ['bb-labeling-agent-trace'],
                            "metadata": {"user_id": self.user_name,
                                         "project": self.langsmith_name}
                        })

        return traced_agent()

    @traceable(run_type="chain", name="Agent Run")
    def run(self, query: str, image: str = None):
        image_base64 = self._image_to_base64(image) if image else self._image_to_base64('./temp_bb_image.jpeg')

        messages = [
            {"role": "system", "content": self.sys_prompt},
            {"role": "user", "content": [
                {"type": "text", "text": query},
                {"type": "image_url", "image_url": f"data:image/jpeg;base64,{image_base64}"}
            ]}
        ]

        response = self.llm.invoke(messages,)
        if self.logger:
            self.logger.debug(f'respone from LLM : {response}')
        if type(response.content)==list:
            response.content= response.content[0]['text'] ## for gemini 3 pro preview model
        raw = response.content.strip()

        if raw.startswith("```"):
            raw = raw.strip("` \n")
            if raw.startswith("json"):
                raw = raw[len("json"):].strip()
        return raw
            
    @traceable(run_type="tool", name="Image to Base64")
    def _image_to_base64(self,image):
        """
        Convert an image file to a base64-encoded string.
        Args:
            image (str): Path to the image file.
        Returns:
            str: Base64-encoded image string.
        """
        if not os.path.exists(image):
            raise FileNotFoundError(f"Image file not found at path: {image}")

        with open(image, "rb") as f:
            return base64.b64encode(f.read()).decode('utf-8')
        

class LabelingState(TypedDict):
    messages: List[Dict[str, Any]]
    boxes_json: Optional[str]
    labeled_objects: Optional[List[Dict[str, Any]]] 
    temp_bb_img_path : Optional[str]
    valid: Optional[bool]
    query: str
    image_path: str
    failed_labels: Optional[List[str]]

class LabelingGraph:
    """
    This graph uses:
    - create_labeling_agent
    - BBTools
    and loops until bounding boxes are validated

    Runing :
    agent = create_labeling_agent(llm)
    graph = LabelingGraph(agent, "image.jpg", "Label all objects.")
    result = graph.run()

    print(result)
    """

    def __init__(self, agent: create_labeling_agent, logger=None):
        self.agent = agent
        self.logger = logger
        self.workflow = self._build_graph()

    @traceable(run_type="chain", name="Labeler Node")
    def node_labeler(self, state: LabelingState):
            """
            Generates/Corrects the bounding box JSON based on the initial query 
            and any feedback from failed_items.
            """
            
            current_query = state["query"]
            if state.get("failed_items"):
                failed_list = ", ".join(state["failed_items"])
                correction_prompt = (
                    f"{current_query}\n\n"
                    f"ATTENTION: The previously generated bounding boxes for the following items were marked as incorrect or missing: **{failed_list}**. "
                    f"Please review the provided image (which shows the last attempt) and **regenerate the complete and correct list of bounding boxes and labels** for all requested items, focusing on correcting the issues for {failed_list}."
                )
            else:
                correction_prompt = current_query
                
            boxes_json = self.agent.run(correction_prompt, state["image_path"])
            
            try:
                parsed_data = json.loads(boxes_json)
                labeled_objects = parsed_data.get("labeled_objects", [])
                if not isinstance(labeled_objects, list):
                    raise TypeError("Expected 'labeled_objects' to be a list.")
            except (json.JSONDecodeError, TypeError) as e:
                msg = f"Warning: LLM returned invalid JSON format: {e}. Forcing re-run."
                if self.logger:
                    self.logger.warning(msg)
                else:
                    print(msg)
                return {
                    **state, 
                    "valid": False,
                    "failed_labels": ["All objects (JSON format error)"]
                }

            return {
                **state, 
                "messages": [{"role": "agent", "content": boxes_json}], 
                "boxes_json": boxes_json,
                "labeled_objects": labeled_objects, # Storing the iterable list
                "failed_labels": None 
            }
    
    @traceable(run_type="tool", name="Bounding Box Visualization Tool")
    def node_tool(self, state: LabelingState):
            """Draws the bounding boxes on the image."""
            tool = BBTools(state['image_path'])
            tool._apply_bounding_boxes(state["boxes_json"], show=True, save_location=state['temp_bb_img_path'])
            return state

    @traceable(run_type="llm", name="Validator Single item LLM Call")
    def _llm_validate_single_item(self, state: LabelingState, item_data: Dict[str, Any]) -> bool:
            """
            Helper to call the LLM to validate a single item.
            The tool must draw ONLY this item's box on a temporary image.
            This is a complex step, as it requires dynamic image generation per loop iteration.
            """
            validation_prompt = f"""
            You are a Bounding Box Validator. Review the following object data and the full image with all boxes drawn on it.
            
            - **Label to Check**: {item_data['label']}
            - **Box Coordinates**: {item_data['box']}
            
            Based on the visual evidence, is the box accurate and tight?
            Your response must be a single JSON object: {{"valid": true}} or {{"valid": false, "reason": "..."}}.
            """
            validation_result_json = self.agent.run(validation_prompt, state['temp_bb_img_path'])
            
            try:
                result = json.loads(validation_result_json)
                return result.get("valid", False)
            except json.JSONDecodeError:
                return False

    @traceable(run_type="chain", name="Validator Node")
    def node_validator(self, state: LabelingState):
        """Iterates over each item and checks its validity."""
        
        updated_objects = []
        all_valid = True
        failed_labels = []

        for item in state.get("labeled_objects", []):
            current_label = item['label']
            is_valid = item.get("valid", False)
            
            if not is_valid:
                break 
        
        validation_result_json = self._llm_validate_full_list(state)
        
        try:
            result = json.loads(validation_result_json)
            all_valid = result.get("valid", False)
            failed_labels = result.get("failed_labels", []) 
        except json.JSONDecodeError:
            all_valid = False
            failed_labels = ["All labels (Validator JSON error)"]
            
        if all_valid:
            msg = "Validation successful."
            if self.logger:
                self.logger.info(msg)
            else:
                print(msg)
            return {**state, "valid": True}
        else:
            msg = f"Validation failed. Items to correct: {failed_labels}"
            if self.logger:
                self.logger.warning(msg)
            else:
                print(msg)
            return {
                **state, 
                "valid": False,
                "failed_labels": failed_labels 
            }
    
    @traceable(run_type="llm", name="Validator LLM Call")
    def _llm_validate_full_list(self, state: LabelingState) -> str:
        """
        Helper to call the LLM for one validation pass on the entire processed image.
        This prompt forces the LLM to return the list of items that failed.
        """
        validation_prompt = f"""
        You are a Bounding Box Validator. Review the image which contains all drawn bounding boxes.
        The objects requested were: {state['query']}
        
        You must evaluate **every single drawn bounding box** and label.
        
        Your response must be a JSON object:
        1. **"valid"**: A boolean (true if ALL boxes/labels are correct, false otherwise).
        2. **"failed_labels"**: A list of strings. If "valid" is true, this list is empty: []. If "valid" is false, list the **names/labels** of all items that are missing, have incorrect bounding boxes, or have incorrect labels (e.g., ["blue chair", "coffee mug (missing)"]).

        Return JSON only.
        """
        return self.agent.run(validation_prompt, state['temp_bb_img_path'])

    
    def route(self, state: LabelingState):
        """Conditional edge: END if valid, otherwise loop back to labeler."""
        return END if state["valid"] else "labeler"

    def _build_graph(self):
        graph = StateGraph(LabelingState)
        graph.add_node("labeler", self.node_labeler)
        graph.add_node("tool", self.node_tool)
        graph.add_node("validator", self.node_validator)

        graph.add_edge(START, "labeler")
        graph.add_edge("labeler", "tool")
        graph.add_edge("tool", "validator")

        graph.add_conditional_edges(
            "validator", self.route, {END: END, "labeler": "labeler"}
        )

        return graph.compile()

    @traceable
    def run(self, image_path: str, query: str, temp_image :str = './data/temp_bb_image.jpg'):
        self.image_path = image_path
        self.query = query
        self.temp_image = temp_image
        return self.workflow.invoke(
            {
                "agent": self.agent,
                "image_path": self.image_path,
                "query": self.query,
                "temp_bb_img_path": self.temp_image,
                "valid": 'false'
            }
        )


## give access of db and get the query with tools

from ..prompts_bank import PromptManager
from langchain.agents import create_agent
import os
from .tools import SQLDatabaseTools
from .middleware import LoggingMiddleware, SQLGuardRailsMiddleware

__all__ = ["runtime_sql_agent", "run_sql_agent"]

SYS_PROMPT = PromptManager().get_template("SQL_AGENT_SYS_PROMPT")

class runtime_sql_agent:
    """
    Setting up the DB connection for SQL agent.

    Example:
        sql_db = runtime_sql_agent(db_connection)
    """

    def __init__(self, db_connection, logger=None):
        self.db_connection = db_connection
        self.logger = logger
    
    def test_basic_mb(self) -> str:
        """
        Test the database connection by executing a simple query.

        Returns:
            str: Result of the test query
        """
        from mb_rag.utils.extra import check_package
        check_package('mb_sql', 'Please install mb_sql package to use test this function: pip install -U mb_sql')
        from mb_sql.utils import list_schemas,list_tables
        try:
            msg = f"{list_schemas(self.db_connection)}\n\n{list_tables(self.db_connection, schema='public')}"
            if self.logger:
                self.logger.info(msg)
            else:
                print(msg)
            return msg
        except Exception as e:
            return f"Database connection failed: {str(e)}"
    
class run_sql_agent:
    """
    Create and return a SQL agent instance.
    
    Args:
        llm: The language model to use.
        db_connection: The database connection object with an `execute_query` method.
        sys_prompt: System prompt for the agent. (Defaults to SYS_PROMPT)
        langsmith_params: If True, enables LangSmith tracing.
        
    Returns:
        Configured SQL agent.
    """

    def __init__(self, 
                 llm,
                db_connection,
                sys_prompt=SYS_PROMPT,
                langsmith_params=True, 
                recursion_limit: int = 50,
                user_name: str = "test_notebook_user",
                logging: bool = False,
                logger=None
                ):

        self.llm = llm
        self.db_connection = db_connection
        self.sys_prompt = sys_prompt
        self.langsmith_params = langsmith_params
        self.recursion_limit = recursion_limit
        self.user_name = user_name
        self.logging = logging
        self.logger = logger
        if self.logging:
            self.middleware = [LoggingMiddleware(logger=self.logger), SQLGuardRailsMiddleware(logger=self.logger)]
        else:
            self.middleware = [SQLGuardRailsMiddleware(logger=self.logger)]

        from mb_rag.utils.extra import check_package
        check_package('mb_sql', 'Please install mb_sql package to use SQL agent with mb_sql: pip install -U mb_sql')
        from mb_sql.sql import read_sql
        from mb_sql.utils import list_schemas
        self.read_sql = read_sql
        self.list_schemas = list_schemas

        if not self.langsmith_params:
            os.environ["LANGCHAIN_TRACING"] = "false"
        else:
            os.environ.setdefault("LANGCHAIN_TRACING", "true")
            self.langsmith_name = os.environ.get("LANGSMITH_PROJECT", "SQL-Agent-Project")
        self.agent = self.create_sql_agent()

    def create_sql_agent(self):
        """
        Create and configure the SQL agent.
        
        Returns:
            Configured SQL agent.
        """

        if self.langsmith_params:
            from langsmith import traceable

            @traceable(run_type="chain", name=self.langsmith_name)
            def traced_agent():
                return create_agent(
                    system_prompt=self.sys_prompt,
                tools=[SQLDatabaseTools(self.db_connection).to_tool_table_info(),
                       SQLDatabaseTools(self.db_connection).to_tool_list_tables(),
                       SQLDatabaseTools(self.db_connection).to_tool_text_to_sql(),
                       SQLDatabaseTools(self.db_connection).to_tool_execute_query(),
                       SQLDatabaseTools(self.db_connection).to_tool_database_schemas()],
                    model=self.llm,
                    context_schema=self.db_connection,
                    middleware=self.middleware,
                ).with_config({"recursion_limit": self.recursion_limit,
                                'tags': ['sql-agent-trace'],
                           "metadata": {"user_id": self.user_name}
                           })

            return traced_agent()
        else:
            # No tracing
            return create_agent(
                system_prompt=self.sys_prompt,
                tools=[SQLDatabaseTools(self.db_connection).to_tool_table_info(),
                       SQLDatabaseTools(self.db_connection).to_tool_list_tables(),
                       SQLDatabaseTools(self.db_connection).to_tool_text_to_sql(),
                       SQLDatabaseTools(self.db_connection).to_tool_execute_query(),
                       SQLDatabaseTools(self.db_connection).to_tool_database_schemas()],
                model=self.llm,
                context_schema=self.db_connection,
                middleware=self.middleware,
            ).with_config({"recursion_limit": self.recursion_limit, 
                           'tags': ['sql-agent-no-trace'],
                           "metadata": {"user_id": self.user_name}
                           })

    def run(self, query: str) -> str:
        """
        Run a SQL query using the configured agent.
        
        Args:
            query: SQL query string.
            
        Returns:
            str: Result of the query execution.
        """
        try:
            for step in self.agent.stream(
                {"messages": [{"role": "user", "content": query}]},
                stream_mode="values",
            ):
                step["messages"][-1].pretty_print()
        except Exception as e:
            msg = f"[Agent Error] {e}"
            if self.logger:
                self.logger.error(msg)
            else:
                print(msg)
            return str(e)

    def _save_to_db():
        """
        Save agent interactions to the given database.

        Returns:
            None
        """
        pass  # Implementation depends on specific database schema and requirements



"""
Web browsing agent implementation using Google's Gemini model
"""

from typing import List, Optional, Dict, Any
from langchain_core.tools import Tool
from langchain_community.tools import DuckDuckGoSearchResults
from langchain_community.utilities import DuckDuckGoSearchAPIWrapper
# from langchain.tools import WebBrowserTools
from mb_rag.basic import ModelFactory
from mb_rag.utils.extra import check_package
# from duckduckgo_search import DDGS
from langchain_core.messages import HumanMessage
from mb_rag.basic import ModelFactory
from typing import Any, Dict

__all__ = [
    'WebBrowserAgent',
    'WebSearchLangGraphAgent'
]

def check_web_dependencies() -> None:
    """
    Check if required web scraping packages are installed
    Raises:
        ImportError: If any required package is missing
    """
    if not check_package("bs4"):
        raise ImportError("BeautifulSoup4 package not found. Please install it using: pip install beautifulsoup4")
    if not check_package("requests"):
        raise ImportError("Requests package not found. Please install it using: pip install requests")


class WebBrowserAgent:
    """
    Agent for web browsing tasks using Google's Gemini model
    
    Attributes:
        model: The Google Generative AI model instance
        search_tool: Search tool for web queries
        browser_tools: Collection of web browsing tools
    """

    def __init__(self, 
                 model_config: dict = {'model_type': 'google','model_name': 'gemini-2.0-flash', 'temperature': 0.7}, 
                 max_results: int = 2,
                 **kwargs):
        """
        Initialize the web browser agent. Currently only duckduckgo is supported.

        Args:
            model_config: Model Configuration for the agent
            max_results: Maximum number of search results to retrieve
            **kwargs: Additional arguments for model initialization
        """
        # Check dependencies before initializing
        check_web_dependencies()
        
        # Initialize configuration
        self.config = model_config 
        self.max_results = max_results

        # Initialize model and tools
        self._initialize_model()
        self._initialize_search()
        self._initialize_browser_tools()
        
        # Import required packages after dependency check
        import requests
        from bs4 import BeautifulSoup
        self._requests = requests
        self._BeautifulSoup = BeautifulSoup

    def _initialize_model(self) -> None:
        """Initialize the AI model"""
        self.model = ModelFactory(**self.config)

    def _initialize_search(self) -> None:
        """Initialize search capabilities"""
        search_wrapper = DuckDuckGoSearchAPIWrapper(max_results=self.max_results)
        self.search_tool = DuckDuckGoSearchResults(api_wrapper=search_wrapper)

    def _initialize_browser_tools(self) -> None:
        """Initialize web browsing tools"""
        self.browser_tools = self._get_browser_tools()

    def _get_browser_tools(self) -> List[Tool]:
        """
        Create and return the set of web browsing tools
        
        Returns:
            List[Tool]: List of web browsing tools
        """
        tools = [
            Tool(
                name="web_search",
                description="Search the web for information using DuckDuckGo",
                func=self.search_tool.run
            ),
            Tool(
                name="fetch_webpage",
                description="Fetch and parse content from a webpage",
                func=self._fetch_webpage
            ),
            Tool(
                name="extract_links",
                description="Extract all links from a webpage",
                func=self._extract_links
            )
        ]
        return tools

    def _fetch_webpage(self, url: str) -> str:
        """
        Fetch and parse content from a webpage
        
        Args:
            url: URL of the webpage to fetch
            
        Returns:
            str: Parsed text content from the webpage
        """
        try:
            response = self._requests.get(url)
            response.raise_for_status()
            soup = self._BeautifulSoup(response.text, 'html.parser')
            
            # Remove script and style elements
            for script in soup(["script", "style"]):
                script.decompose()
                
            # Get text content
            text = soup.get_text()
            
            # Clean up text
            lines = (line.strip() for line in text.splitlines())
            chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
            text = ' '.join(chunk for chunk in chunks if chunk)
            
            return text
        except Exception as e:
            return f"Error fetching webpage: {str(e)}"

    def _extract_links(self, url: str) -> List[str]:
        """
        Extract all links from a webpage
        
        Args:
            url: URL of the webpage to extract links from
            
        Returns:
            List[str]: List of URLs found on the webpage
        """
        try:
            response = self._requests.get(url)
            response.raise_for_status()
            soup = self._BeautifulSoup(response.text, 'html.parser')
            
            links = []
            for link in soup.find_all('a'):
                href = link.get('href')
                if href and href.startswith('http'):
                    links.append(href)
                    
            return links
        except Exception as e:
            return [f"Error extracting links: {str(e)}"]

    def browse(self, query: str) -> str:
        """
        Execute a web browsing task based on the query
        
        Args:
            query: The search query or URL to process
            
        Returns:
            str: Results from the web browsing task
        """
        if self._is_url(query):
            return self._fetch_webpage(query)
        return self.search_tool.run(query)

    @staticmethod
    def _is_url(text: str) -> bool:
        """
        Check if text is a URL
        
        Args:
            text: Text to check
            
        Returns:
            bool: True if text is a URL
        """
        return text.startswith(('http://', 'https://'))

    @property
    def tools(self) -> List[Tool]:
        """Get all available web browsing tools"""
        return self.browser_tools

    def get_tool(self, name: str) -> Optional[Tool]:
        """
        Get a specific tool by name
        
        Args:
            name: Name of the tool
            
        Returns:
            Optional[Tool]: The requested tool or None if not found
        """
        return next((tool for tool in self.tools if tool.name == name), None)


class WebSearchLangGraphAgent:
    """
    A LangGraph Agent for performing web searches and returning a structured JSON summary.

    Example:
        agent = WebSearchLangGraphAgent(model_name="gpt-4o")
        output = agent.search("Latest developments in renewable energy")
        if self.logger:
            self.logger.info(output)
        else:
            print(output)
    """

    def __init__(self, model_name: str = "gpt-4o", logger=None, **kwargs):
        # Initialize the underlying model using ModelFactory
        self.llm = ModelFactory(model_name=model_name, **kwargs).model
        self.logger = logger
        
        from langgraph.graph import MessageGraph
        self.graph = MessageGraph()

        # Add nodes
        self.graph.add_node("user", self._user_node)
        self.graph.add_node("search", self._search_node)
        self.graph.add_node("summarize", self._summarize_node)

        # Set the workflow: user -> search -> summarize
        self.graph.add_edge("user", "search")
        self.graph.add_edge("search", "summarize")

        # Set the entry point of the agent to the user node
        self.graph.set_entry_point("user")

        # Compile the graph into an executable agent
        self.compiled_agent = self.graph.compile()


    def _user_node(self, state: Dict[str, Any], input_text: str) -> Dict[str, Any]:
        """
        Node to capture the user's search query.
        """
        state["query"] = input_text
        return state

    def _search_node(self, state: Dict[str, Any]) -> Dict[str, Any]:
        """
        Node to perform the web search.
        """
        query = state.get("query", "")
        raw_results = WebSearchLangGraphAgent.perform_web_search(query)
        state["raw_results"] = raw_results
        return state

    def _summarize_node(self, state: Dict[str, Any]) -> Dict[str, Any]:
        """
        Node that uses the LLM to generate a structured JSON summary of the web search results.
        """
        raw_results = state.get("raw_results", "")
        prompt = (
            "You are a helpful AI assistant. Please analyze and summarize the web search results below. "
            "Return a structured JSON with two keys: 'summary' for a concise overview and 'details' for more in-depth information.\n\n"
            f"Results:\n{raw_results}"
        )
        message = HumanMessage(content=prompt)
        response = self.llm.invoke([message])
        try:
            state["structured_output"] = response.content
        except Exception:
            state["structured_output"] = response
        return state

    def search(self, query: str) -> str:
        """
        Perform a web search and obtain a structured JSON summary.
        
        Args:
            query (str): The search query.
        
        Returns:
            str: The structured JSON summary generated by the LLM.
        """
        state: Dict[str, Any] = {}
        state = self.compiled_agent(query, state)
        return state.get("structured_output", "No structured output produced")


## simple llm wrapper to replace invoke with invoke_query/own model query

__all__ = ["LLMWrapper"]

class LLMWrapper:
    """A simple wrapper for the language model to standardize the invoke method.
    """

    def __init__(self, llm):
        self.llm = llm

    def __getattr__(self, name):
        """Get all attributes from llm module. (invoke_query, invoke_query_threads, etc.)"""
        return getattr(self.llm, name)

    def invoke(self, use_threads=False,**kwargs) -> str:
        """
        Invoke the language model with a list of messages.
        Using invoke_query method of the underlying model.
        Check ModelFactory for more details.
        
        Args:
            use_threads (bool): Whether to use threading for invocation. Defaults to False.
            **kwargs: Keyword arguments for the model invocation.

        Returns:
            str: The generated response.
        """
        if use_threads:
            return self.llm.invoke_query_threads(**kwargs)
        return self.llm.invoke_query(**kwargs)